------------------------------------------
-- Setting up libraries
------------------------------------------

local AddonName = "iFriends"
iFriends = LibStub("AceAddon-3.0"):NewAddon(AddonName, "AceEvent-3.0", "AceTimer-3.0");

local QTip = LibStub("LibQTip-1.0");
local Tourist = LibStub("LibTourist-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale(AddonName);

local Tooltip;

iFriends.Broker = LibStub("LibDataBroker-1.1"):NewDataObject(AddonName, {
	type = "data source",
	text = "Loading...",
	icon = "Interface\\Addons\\iFriends\\Fubar_FriendsFu",

	OnClick = function(_, button)
		-- if we click while ALT is hold, we add a new friend
		-- thanks to blizz for that code :)
		if( button == "LeftButton" ) then
			if( IsAltKeyDown() ) then
				if( BNFeaturesEnabled() ) then
					AddFriendEntryFrame_Collapse(true);
					AddFriendFrame.editFocus = AddFriendNameEditBox;
					StaticPopupSpecial_Show(AddFriendFrame);
					if( GetCVarBool("addFriendInfoShown") ) then
						AddFriendFrame_ShowEntry();
					else
						AddFriendFrame_ShowInfo();
					end
				else
					StaticPopup_Show("ADD_FRIEND");
				end
			else
				ToggleFriendsFrame(1);
			end
		elseif( button == "RightButton" ) then
			iFriends:OpenOptions();
		end
	end,

	OnEnter = function(anchor)
		ShowFriends();
		Tooltip = QTip:Acquire(iFriends);
		Tooltip:SetAutoHideDelay(0.01, anchor);
		Tooltip:SmartAnchorTo(anchor);
		iFriends:UpdateTooltip();
		Tooltip:Show();
	end,
});

------------------------------------------
-- Variables and colors
------------------------------------------

local NULL = "";
local COLUMN_HEADER_ROW = 0;

local COLOR_GOLD = "|cfffed100";
local COLOR_BNET = FRIENDS_BNET_NAME_COLOR_CODE;

local ClassColors = {};
for k, v in pairs(LOCALIZED_CLASS_NAMES_MALE) do
	local c = RAID_CLASS_COLORS[k];
	ClassColors[v] = ("|cff%02x%02x%02x"):format(c.r *255, c.g *255, c.b *255);
end
for k, v in pairs(LOCALIZED_CLASS_NAMES_FEMALE) do
	local c = RAID_CLASS_COLORS[k];
	ClassColors[v] = ("|cff%02x%02x%02x"):format(c.r *255, c.g *255, c.b *255);
end

------------------------------------------
-- Sorting and Columns
------------------------------------------

iFriends.Sort = {
	name = function(a, b)
		return a.charName < b.charName;
	end,
};

iFriends.Columns = {
	realid = {
		Label = L["RealID"],
		Brush = function(v)
			return ("%s%s|r"):format(COLOR_BNET, v.realID or NULL);
		end,
	},
	realm = {
		Label = L["Game/Realm"],
		Brush = function(v)
			if( v.Game == BNET_CLIENT_WOW ) then
				if( v.BattleNet ) then
					local factioncity = {L["Hordecity"], L["Allycity"]};
					local r, g, b = Tourist:GetFactionColor(factioncity[v.charFaction +1]);
					return ("|cff%02x%02x%02x%s|r"):format(r *255, g *255, b *255, v.charRealm);
				else
					return ("%s%s|r"):format(FRIENDS_OTHER_NAME_COLOR_CODE, L["local friend"]);
				end
			else
				if( v.Game == BNET_CLIENT_SC2 ) then
					return "Starcraft 2";
				elseif( v.Game == "D3" ) then
					return "Diablo 3";
				else
					return v.Game;
				end
			end
		end,
	},
	level = {
		Label = L["Level"],
		Brush = function(v)
			if( v.Game == BNET_CLIENT_WOW ) then
				local c = GetQuestDifficultyColor(v.charLevel);
				return ("|cff%02x%02x%02x%s|r"):format(c.r *255, c.g *255, c.b *255, v.charLevel);
			else
				return NULL;
			end
		end,
	},
	name = {
		Label = L["Name"],
		Brush = function(v)
			if( v.Game == BNET_CLIENT_WOW ) then
				return ("%s%s%s|r"):format(ClassColors[v.charClass], v.charStatus or NULL, v.charName);
			else
				return NULL;
			end
		end,
	},
	race = {
		Label = L["Race"],
		Brush = function(v)
			if( v.Game == BNET_CLIENT_WOW and v.BattleNet ) then
				local factioncity = {L["Hordecity"], L["Allycity"]};
				local r, g, b = Tourist:GetFactionColor(factioncity[v.charFaction +1]);
				return ("|cff%02x%02x%02x%s|r"):format(r *255, g *255, b *255, v.charRace);
			else
				return NULL;
			end
		end,
	},
	zone = {
		Label = L["Zone"],
		Brush = function(v)
			if( v.Game == BNET_CLIENT_WOW ) then
				local r, g, b = Tourist:GetFactionColor(v.charZone);
				return ("|cff%02x%02x%02x%s|r"):format(r *255, g *255, b *255, v.charZone);
			else
				return NULL;
			end
		end,
	},
	note = {
		Label = L["Note"],
		Brush = function(v)
			if( v.Broadcast and v.Broadcast ~= "" ) then
				return ("%s%s|r"):format(COLOR_BNET, v.Broadcast);
			elseif( v.localNote and v.localNote ~= "" ) then
				return ("%s%s|r"):format(COLOR_GOLD, v.localNote);
			end
			
			return NULL;
		end,
	},
};

------------------------------------------
-- GetColumns
------------------------------------------

function iFriends:GetColumns()
	local columns = self.db.Display;
	local new = {};

	for i = 1, #columns do
		local insert = true;
		local canUse = self.Columns[columns[i]].CanUse;
		
		if( type(canUse) == "boolean" and not canUse ) then
			insert = false;
		elseif( type(canUse) == "function" and not canUse() ) then
			insert = false;
		end
		
		if( insert ) then
			table.insert(new, columns[i]);
		end
	end
	
	return new;
end

------------------------------------------
-- OnInitialize
------------------------------------------

function iFriends:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("iFriendsDB", self:CreateDB(), "Default").profile;
	
	self.Friends = {};

	self:RegisterEvent("FRIENDLIST_UPDATE", "EventHandler");
	self:RegisterEvent("BN_FRIEND_INFO_CHANGED", "EventHandler");
	self:RegisterEvent("BN_FRIEND_LIST_SIZE_CHANGED", "EventHandler");
	
	ShowFriends();
	self:ScheduleRepeatingTimer(ShowFriends, 55);
end

------------------------------------------
-- EventHandler
------------------------------------------

function iFriends:EventHandler(event)
	local TotalF, TotalFOn = GetNumFriends();
	local TotalBN, TotalBNOn = 0, 0;
	if( BNFeaturesEnabledAndConnected() ) then
		TotalBN, TotalBNOn = BNGetNumFriends();
	end
	
	local Total = TotalF + TotalBN;
	local TotalOn = TotalFOn + TotalBNOn;
	
	self.Broker.text = ("%d/%d"):format(TotalOn, Total);
	self:SetupFriendsData(TotalFOn, TotalBNOn);
	
	if( QTip:IsAcquired(self) ) then
		self:UpdateTooltip();
	end
end

------------------------------------------
-- SetupFriendsData
------------------------------------------

function iFriends:SetupFriendsData(TotalFOn, TotalBNOn)
	local list = {};
	
	for i = 1, TotalFOn do
		local charName, charLevel, charClass, charZone, isOnline, charStatus, localNote = GetFriendInfo(i);
		
		if( isOnline ) then
			local vartable = {
				BattleNet = false,
				Game = BNET_CLIENT_WOW,
				charName = charName,
				charLevel = charLevel,
				charClass = charClass,
				charZone = charZone,
				charStatus = charStatus or nil,
				localNote = localNote or nil,
			};
			
			list[i] = vartable;
		end
	end
	
	if( BNFeaturesEnabledAndConnected() ) then
		for i = 1, TotalBNOn do
			local pID, givenName, surName, _, _, _, isOnline, _, isAFK, isDND, Broadcast, localNote, _, _ = BNGetFriendInfo(i);
			
			if( isOnline ) then
				_, charName, Game, charRealm, _, charFaction, charRace, charClass, _, charZone, charLevel, _ = BNGetToonInfo(pID);
				
				local realID = BATTLENET_NAME_FORMAT:format(givenName, surName);
				
				local charStatus = nil;
				if( isAFK ) then
					charStatus = ("<%s>"):format(AFK);
				elseif( isDND ) then
					charStatus = ("<%s>"):format(DND);
				end
				
				local vartable = {
					BattleNet = true,
					Game = Game,
					pID = pID,
					realID = realID,
					charName = charName,
					charLevel = charLevel,
					charClass = charClass,
					charZone = charZone,
					charStatus = charStatus or nil,
					localNote = localNote or nil,
					charRealm = charRealm,
					charRace = charRace,
					charFaction = charFaction,
					Broadcast = Broadcast or nil,
				};
				
				list[TotalFOn + i] = vartable;
			end
		end
	end
	
	--table.sort(list, self.Sort[self.db.Sort]);
	self.Friends = list;
end

------------------------------------------
-- OnClick callbacks
------------------------------------------

local function cbTooltipLineClick(_, v, button)
	if( v.BattleNet ) then
		local itemRef = ("%s:%d"):format(v.realID, v.pID)
		SetItemRef(("BNplayer:%s"):format(itemRef), ("|HBNplayer:%s|h[%s]|h"):format(itemRef, v.realID), "LeftButton")
	else
		if( IsAltKeyDown() ) then
			InviteUnit(v.charName);
		else
			SetItemRef(("player:%s"):format(v.charName), ("|Hplayer:%s|h[%s]|h"):format(v.charName, v.charName), "LeftButton");
		end
	end
end

------------------------------------------
-- UpdateTooltip
------------------------------------------

function iFriends:UpdateTooltip()
	Tooltip:Clear();
	
	local ActiveCols = self:GetColumns();
	Tooltip:SetColumnLayout(#ActiveCols);
	
	for NumLine = COLUMN_HEADER_ROW, #self.Friends do
		local friend;

		for NumCol = 1, #ActiveCols do
			local ColName = ActiveCols[NumCol];
			local ColInfo = self.Columns[ColName];
			
			if( NumCol == 1 ) then
				if( NumLine == COLUMN_HEADER_ROW) then
					Tooltip:AddHeader(" ");
				else
					Tooltip:AddLine(" ");
				end
			end
			
			if( NumLine == COLUMN_HEADER_ROW ) then
				local label = "";
				if( self.db.Column[ColName].ShowLabel ) then
					label = ColInfo.Label;
				end
				
				Tooltip:SetCell(NumLine +1, NumCol, label, nil, self.db.Column[ColName].Align);
			else
				friend = self.Friends[NumLine];
				
				Tooltip:SetCell(NumLine +1, NumCol, ColInfo.Brush(friend), nil, self.db.Column[ColName].Align);
				if( ColInfo.Script and ColInfo.ScriptUse() ) then
					Tooltip:SetCellScript(NumLine +1, NumCol, "OnMouseDown", ColInfo.Script, friend);
				end
				
			end
		end
		
		if( friend ) then
			Tooltip:SetLineScript(NumLine +1, "OnMouseDown", cbTooltipLineClick, friend);
			
			if( friend.Game ~= BNET_CLIENT_WOW ) then
				local r, g, b = 1, 1, 1;
				
				if( friend.Game == BNET_CLIENT_SC2 ) then
					r, g, b = 0.1, 0.8, 1;
				elseif( friend.Game == "D3") then
					r, g, b = 1, 0.1, 0.1;
				end
				
				Tooltip:SetLineColor(NumLine +1, r, g, b, 0.6);
			end
		end
	end
end